# coding=utf-8

from __future__ import unicode_literals

from spacy.symbols import (  # type: ignore
    ADJ,
    ADP,
    ADV,
    AUX,
    CCONJ,
    DET,
    INTJ,
    NOUN,
    NUM,
    PART,
    POS,
    PRON,
    PROPN,
    PUNCT,
    SPACE,
    SYM,
    VERB,
)

TAG_MAP = {
    "名詞,普通名詞": {POS: NOUN},
    "名詞,サ変名詞": {POS: NOUN},
    "名詞,固有名詞": {POS: PROPN},
    "名詞,地名": {POS: PROPN},
    "名詞,人名": {POS: PROPN},
    "名詞,組織名": {POS: PROPN},
    "名詞,数詞": {POS: NUM},
    "名詞,形式名詞": {POS: NOUN},
    "名詞,副詞的名詞": {POS: NOUN},
    "名詞,時相名詞": {POS: NOUN},
    "指示詞,名詞形態指示詞": {POS: PRON},
    "指示詞,連体詞形態指示詞": {POS: DET},
    "指示詞,副詞形態指示詞": {POS: ADV},
    "動詞,*": {POS: VERB},
    "形容詞,*": {POS: ADJ},
    "判定詞,*": {POS: AUX},
    "助動詞,*": {POS: AUX},
    "副詞,*": {POS: ADV},
    "助詞,格助詞": {POS: ADP},
    "助詞,副助詞": {POS: ADP},
    "助詞,接続助詞": {POS: ADP},
    "助詞,終助詞": {POS: PART},
    "接続詞,*": {POS: CCONJ},
    "連体詞,*": {POS: DET},
    "感動詞,*": {POS: INTJ},
    "接頭辞,名詞接頭辞": {POS: NOUN},
    "接頭辞,動詞接頭辞": {POS: ADV},
    "接頭辞,イ形容詞接頭辞": {POS: PART},
    "接頭辞,ナ形容詞接頭辞": {POS: PART},
    "接尾辞,名詞性名詞接尾辞": {POS: NOUN},
    "接尾辞,名詞性述語接尾辞": {POS: NOUN},
    "接尾辞,名詞性名詞助数辞": {POS: NOUN},
    "接尾辞,名詞性特殊接尾辞": {POS: NOUN},
    "接尾辞,形容詞性述語接尾辞": {POS: AUX},
    "接尾辞,形容詞性名詞接尾辞": {POS: PART},
    "接尾辞,動詞性接尾辞": {POS: AUX},
    "特殊,句点": {POS: PUNCT},
    "特殊,読点": {POS: PUNCT},
    "特殊,括弧始": {POS: PUNCT},
    "特殊,括弧終": {POS: PUNCT},
    "特殊,記号": {POS: PUNCT},
    "特殊,空白": {POS: SPACE},
    "未定義語,カタカナ": {POS: NOUN},
    "未定義語,アルファベット": {POS: SYM},
    "未定義語,その他": {POS: SYM},
}



"""Defines utilities for pytorch."""
import contextlib
import operator
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Iterator, Optional, Sequence, Union, cast

import torch
import torch.nn as nn
from spacy.pipeline import Pipe
from spacy.tokens import Doc
from torch._C import is_grad_enabled  # type: ignore

# the type torch.optim.Optimizer uses
OptimizerParameters = Union[Iterable[torch.Tensor], Iterable[Dict[str, Any]]]


class TorchPipe(Pipe):
    def __init__(self, vocab, model=True, **cfg):
        self.vocab = vocab
        self.model = model
        self._device = torch.device("cpu")
        self.cfg = cfg

    @property
    def device(self):
        if not hasattr(self, "_device"):
            self._device = torch.device("cpu")
        return self._device

    def to(self, device: torch.device):
        self._device = device
        if self.model and not isinstance(self.model, bool):
            self.model.to(device)

    def optim_parameters(self) -> OptimizerParameters:
        """Return parameters to be optimized."""
        self.require_model()
        if self.cfg.get("freeze"):
            return []
        return cast(nn.Module, self.model).parameters()


@dataclass
class TensorWrapper:
    """Pytorch tensor wrapper for efficient handling of part of batch tensors in spacy pipeline"""

    batch_tensor: torch.Tensor
    i: int
    length: Optional[int] = None

    def get(self) -> torch.Tensor:
        if self.length is not None:
            return self.batch_tensor[self.i, : self.length]
        return self.batch_tensor[self.i]


GoldCat = Dict[str, float]


def goldcat_to_label(goldcat: GoldCat) -> str:
    assert len(goldcat)
    return max(goldcat.items(), key=operator.itemgetter(1))[0]


def goldcats_to_tensor(
    goldcats: Iterable[GoldCat], label2id: Dict[str, int]
) -> torch.Tensor:
    ids = [label2id[goldcat_to_label(cat)] for cat in goldcats]
    return torch.tensor(ids)


TORCH_LOSS = "torch_loss"


def get_loss_from_docs(docs: Iterable[Doc]) -> torch.Tensor:
    _losses = (doc.user_data.get(TORCH_LOSS) for doc in docs)
    losses = [loss for loss in _losses if isinstance(loss, torch.Tensor)]
    if not losses:
        raise ValueError("loss is not set to docs.")
    tlosses = torch.stack(losses)
    return torch.sum(tlosses)


def add_loss_to_docs(docs: Sequence[Doc], loss: torch.Tensor):
    """Add loss to docs' existing loss. """
    doc = docs[0]
    if TORCH_LOSS in doc.user_data:
        doc.user_data[TORCH_LOSS] += loss
    else:
        doc.user_data[TORCH_LOSS] = loss


@contextlib.contextmanager
def set_grad(grad: bool) -> Iterator[None]:
    prev = is_grad_enabled()
    torch.set_grad_enabled(grad)
    yield
    torch.set_grad_enabled(prev)



import os
import shutil
from pathlib import Path
from typing import Optional, Type

import sentencepiece as spm
from spacy.language import Language
from spacy.tokens import Doc


class EXTS:
    """For spacy.Underscore"""

    pieces_ = "spm_pieces_"


def install_extensions():
    Doc.set_extension(EXTS.pieces_, default=None, force=True)


class Tokenizer:
    SPACE_CHAR = "▁"
    SPIECE_MODEL = "spiece.model"

    def __init__(
        self,
        cls: Type["Defaults"],
        nlp: Optional[Language] = None,
        model_path: str = "",
    ):
        self.vocab = nlp.vocab if nlp is not None else cls.create_vocab(nlp)
        self.tokenizer = spm.SentencePieceProcessor()
        self.model_path = model_path

    def __call__(self, text: str) -> Doc:
        _tokens = self.tokenizer.EncodeAsPieces(text)
        spaces = [
            True if next_token.startswith(self.SPACE_CHAR) else False
            for token, next_token in zip(_tokens, _tokens[1:])
            if token != self.SPACE_CHAR
        ] + [False]
        tokens = [
            token.lstrip(self.SPACE_CHAR)
            for token in _tokens
            if token != self.SPACE_CHAR
        ]
        doc = Doc(self.vocab, tokens, spaces)
        doc._.set(EXTS.pieces_, _tokens)
        return doc

    def load_spm_tokenizer(self):
        if os.path.isdir(self.model_path):
            self.model_path = os.path.join(self.model_path, self.SPIECE_MODEL)
        try:
            self.tokenizer.load(self.model_path)
        except OSError:
            pass

    @property
    def model_path(self):
        return self._model_path

    @model_path.setter
    def model_path(self, model_path: str):
        """Set model_path automatically reload sentencepiece tokenizer"""
        self._model_path = model_path
        if model_path:
            self.load_spm_tokenizer()

    def to_disk(self, path: Path, **kwargs):
        path.mkdir(exist_ok=True)
        if self.model_path:
            shutil.copy(self.model_path, path / self.SPIECE_MODEL)

    def from_disk(self, path: Path, **kwargs):
        self.model_path = str((path / self.SPIECE_MODEL).absolute())


class Defaults(Language.Defaults):  # type: ignore
    lex_attr_getters = dict(Language.Defaults.lex_attr_getters)

    @classmethod
    def create_tokenizer(cls, nlp=None, model_path: str = ""):
        return Tokenizer(cls, nlp, model_path=model_path)


class SentencePieceLang(Language):
    lang = "sentencepiece"
    Defaults = Defaults

    def make_doc(self, text: str) -> Doc:
        return self.tokenizer(text)


install_extensions()
Language.factories[SentencePieceLang.lang] = SentencePieceLang

__all__ = ["SentencePieceLang"]



"""Elmo: Deep contextualized word representations (https://allennlp.org/elmo)"""
import json
import shutil
from pathlib import Path
from typing import Generator, Iterable, List

import numpy as np
import spacy
import torch
from spacy.pipeline import Pipe
from spacy.tokens import Doc

from camphr.pipelines.allennlp_base import Pathlike
from camphr.pipelines.utils import (
    get_doc_vector_via_tensor,
    get_similarity,
    get_span_vector_via_tensor,
    get_token_vector_via_tensor,
)


@spacy.component("elmo", assigns=["doc.tensor", "doc.vector", "token.vector"])
class Elmo(Pipe):
    WEIGHTS_FILE_NAME = "weights.hdf5"
    OPTIONS_FILE_NAME = "options.json"

    def __init__(self, model=None, **cfg):
        self.model = model
        self.cfg = cfg

    @classmethod
    def from_elmofiles(cls, options_file: Pathlike, weight_file: Pathlike, **cfg):
        """Construct Elmo with elmo files.
        
        Note:
            Elmo files are distributed in https://allennlp.org/elmo
        
        Example:
            >>> elmo = Elmo.from_elmofiles("./options.json", "./weights.hd5")
            >>> nlp = spacy.blank("en")
            >>> nlp.add_pipe(elmo)
        """
        model = cls.Model(options_file, weight_file)
        return cls(model, **cfg)

    @classmethod
    def Model(cls, options_file: Pathlike, weight_file: Pathlike, **cfg):
        from allennlp.modules.elmo import Elmo

        return Elmo(
            str(Path(options_file).absolute()), str(Path(weight_file).absolute()), 3
        )

    def to_disk(self, path: Pathlike, **cfg):
        path = Path(path)
        path.mkdir(exist_ok=True)
        shutil.copy(
            self.model._elmo_lstm._token_embedder._weight_file,
            str(path / self.WEIGHTS_FILE_NAME),
        )
        with (path / self.OPTIONS_FILE_NAME).open("w") as f:
            json.dump(self.model._elmo_lstm._token_embedder._options, f)

    def from_disk(self, path: Pathlike, **cfg):
        path = Path(path)
        self.model = self.Model(
            path / self.OPTIONS_FILE_NAME, path / self.WEIGHTS_FILE_NAME
        )

    def predict(self, docs: List[Doc]) -> Generator[np.ndarray, None, None]:
        from allennlp.modules.elmo import batch_to_ids

        tokens = [[token.text for token in doc] for doc in docs]
        return self.model(batch_to_ids(tokens))["elmo_representations"]

    def set_annotations(self, docs: List[Doc], outputs: Iterable[torch.Tensor]):
        for doc, vec in zip(docs, outputs):
            vec = vec.detach().cpu().numpy()
            assert vec.shape[1] == len(doc), f"vector: {vec.shape}, doc: {len(doc)}"
            doc.tensor = vec[-1]
            doc.user_hooks["vector"] = get_doc_vector_via_tensor
            doc.user_span_hooks["vector"] = get_span_vector_via_tensor
            doc.user_token_hooks["vector"] = get_token_vector_via_tensor
            doc.user_hooks["similarity"] = get_similarity
            doc.user_span_hooks["similarity"] = get_similarity
            doc.user_token_hooks["similarity"] = get_similarity
            doc.user_data["elmo"] = vec



"""The utils module defines util functions used accross sub packages."""
import bisect
import distutils.spawn
import importlib
import re
from collections import OrderedDict
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import spacy
import srsly
import yaml
from more_itertools import padded
from omegaconf import Config, OmegaConf
from spacy.errors import Errors
from spacy.language import BaseDefaults
from spacy.tokens import Doc, Span, Token
from spacy.util import filter_spans
from toolz import curry
from typing_extensions import Literal

from camphr.types import Pathlike
from camphr.VERSION import __version__


def zero_pad(a: List[List[int]], pad_value: int = 0) -> List[List[int]]:
    """Padding the input so that the lengths of the inside lists are all equal."""
    if len(a) == 0:
        return []
    max_length = max(len(el) for el in a)
    if max_length == 0:
        return a
    return [list(padded(el, fillvalue=pad_value, n=max_length)) for el in a]


RE_URL = re.compile(
    r"https?://(www\.)?[-a-zA-Z0-9@:%._+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&/=]*)"
)


def token_from_char_pos(doc: Doc, i: int) -> Token:
    token_idxs = [t.idx for t in doc]
    return doc[bisect.bisect(token_idxs, i) - 1]


def destruct_token(doc: Doc, *char_pos: int) -> Doc:
    for i in char_pos:
        with doc.retokenize() as retokenizer:
            token = token_from_char_pos(doc, i)
            heads = [token] * len(token)
            retokenizer.split(doc[token.i], list(token.text), heads=heads)
    return doc


def get_doc_char_span(
    doc: Doc, i: int, j: int, destructive: bool = True, **kwargs
) -> Optional[Span]:
    """Get Span from Doc with char position, similar to doc.char_span.

    Args:
        i: The index of the first character of the span
        j: The index of the first character after the span
        destructive: If True, tokens in [i,j) will be splitted and make sure to return span.
        kwargs: passed to Doc.char_span
    """
    span = doc.char_span(i, j, **kwargs)
    if not span and destructive:
        destruct_token(doc, i, j)
        span = doc.char_span(i, j, **kwargs)
    return span


def get_doc_char_spans_list(
    doc: Doc, spans: Iterable[Tuple[int, int]], destructive: bool = True, **kwargs
) -> List[Span]:
    res = []
    for i, j in spans:
        span = get_doc_char_span(doc, i, j, destructive=destructive, **kwargs)
        if span:
            res.append(span)
    return res


def merge_spans(doc: Doc, spans: Iterable[Span]):
    spans = filter_spans(spans)
    with doc.retokenize() as retokenizer:
        for span in spans:
            retokenizer.merge(span)


def split_keepsep(text: str, sep: str):
    texts = text.split(sep)
    if len(texts) == 1:
        return [text]

    res = [t + sep for t in texts[:-1]]
    last = texts[-1]
    if len(last):
        if text.endswith(sep):
            last += sep
        res.append(last)
    return res


def import_attr(import_path: str) -> Any:
    items = import_path.split(".")
    module_name = ".".join(items[:-1])
    return getattr(importlib.import_module(module_name), items[-1])


def get_requirements_line():
    return f"camphr>={__version__}"


def get_defaults(lang: str) -> Type[BaseDefaults]:
    try:
        lang_cls = spacy.util.get_lang_class(lang)
    except Exception:
        return BaseDefaults
    return getattr(lang_cls, "Defaults", BaseDefaults)


def get_labels(labels_or_path: Union[List[str], Pathlike]) -> List[str]:
    if isinstance(labels_or_path, (str, Path)):
        path = Path(labels_or_path)
        if path.suffix == ".json":
            return srsly.read_json(labels_or_path)
        elif path.suffix in {".yml", ".yaml"}:
            return yaml.safe_load(path.read_text())
    return cast(List[str], labels_or_path)


def get_by_dotkey(d: dict, dotkey: str) -> Any:
    assert dotkey
    keys = dotkey.split(".")
    cur = d
    for key in keys:
        assert hasattr(cur, "get"), f"Try to load '{dotkey}' from `{d}`, but not found."
        cur = cur.get(key, None)
        if cur is None:
            return None
    return cur


def create_dict_from_dotkey(dotkey: str, value: Any) -> Dict[str, Any]:
    assert dotkey
    keys = dotkey.split(".")
    result: Dict[str, Any] = {}
    cur = result
    for key in keys[:-1]:
        cur[key] = {}
        cur = cur[key]
    cur[keys[-1]] = value
    return result


@curry
def resolve_alias(aliases: Dict[str, str], cfg: Config) -> Config:
    for alias, name in aliases.items():
        v = get_by_dotkey(cfg, alias)
        if v is None:
            continue
        cfg = OmegaConf.merge(cfg, OmegaConf.create(create_dict_from_dotkey(name, v)))
    return cfg


"""
The following `SerializationMixin` is adopted from spacy-transformers,
which is distributed under the following license:

MIT License

Copyright (c) 2019 ExplosionAI GmbH

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class SerializationMixin:
    """Serializes the items in `serialization_fields`

    Example:
        >>> class FooComponent(SerializationMixin, Pipe):
        >>>     serialization_fields = ["bar_attribute"]
        >>> comp = FooComponent(Vocab())
        >>> save_dir = Path("baz_directory")
        >>> comp.to_disk(save_dir) # saved the component into directory
        >>> loaded_comp = spacy.from_disk(save_dir) # load from directory
    """

    serialization_fields: List[str] = []
    name: str

    def from_bytes(self, bytes_data, **kwargs):
        pkls = srsly.pickle_loads(bytes_data)
        for field in self.serialization_fields:
            setattr(self, field, pkls[field])
        return self

    def to_bytes(self, **kwargs):
        pkls = OrderedDict()
        for field in self.serialization_fields:
            pkls[field] = getattr(self, field, None)
        return srsly.pickle_dumps(pkls)

    def from_disk(self, path: Path, **kwargs):
        path.mkdir(exist_ok=True)
        with (path / "data.pkl").open("rb") as file_:
            data = file_.read()
        return self.from_bytes(data, **kwargs)

    def to_disk(self, path: Path, **kwargs):
        path.mkdir(exist_ok=True)
        data = self.to_bytes(**kwargs)
        with (path / "data.pkl").open("wb") as file_:
            file_.write(data)

    def require_model(self):
        if getattr(self, "model", None) in (None, True, False):
            raise ValueError(Errors.E109.format(name=self.name))


T = TypeVar("T")


def _setdefault(obj: Any, k: str, v: T) -> T:
    """Set attribute to object like dict.setdefault."""
    if hasattr(obj, k):
        return getattr(obj, k)
    setattr(obj, k, v)
    return v


def setdefaults(obj: Any, kv: Dict[str, Any]):
    """Set all attribute in kv to object"""
    for k, v in kv.items():
        _setdefault(obj, k, v)


def get_juman_command() -> Optional[Literal["juman", "jumanpp"]]:
    for cmd in ["jumanpp", "juman"]:
        if distutils.spawn.find_executable(cmd):
            return cmd  # type: ignore
    return None



"""Module trf_model defines pytorch-transformers components."""
import dataclasses
from typing import Any, List, Optional, Tuple

import numpy as np
import spacy
import spacy.language
import torch
import transformers
from spacy.gold import GoldParse
from spacy.tokens import Doc

from camphr.pipelines.utils import get_similarity
from camphr.torch_utils import TensorWrapper, TorchPipe, set_grad

from .auto import get_trf_model_cls
from .tokenizer import TrfTokenizer
from .utils import ATTRS, TrfAutoMixin

spacy.language.ENABLE_PIPELINE_ANALYSIS = True


@dataclasses.dataclass
class TrfModelInputs:
    input_ids: torch.Tensor
    token_type_ids: Optional[torch.Tensor] = None
    attention_mask: Optional[torch.Tensor] = None


TRANSFORMERS_MODEL = "transformers_model"


@spacy.component(TRANSFORMERS_MODEL, assigns=[f"doc._.{ATTRS.last_hidden_state}"])
class TrfModel(TrfAutoMixin[transformers.PreTrainedModel], TorchPipe):
    """Transformers Model component."""

    _TRF_NAME = "trf_name"
    _MODEL_CLS_GETTER = get_trf_model_cls

    @property
    def max_length(self) -> int:
        return self.model.config.max_position_embeddings

    def predict(self, docs: List[Doc]) -> torch.Tensor:
        self.model.eval()
        return self._apply_model(docs, False)

    def _apply_model(self, docs: List[Doc], grad: bool) -> torch.Tensor:
        self.require_model()
        x = TrfTokenizer.get_transformers_input(docs)
        assert x is not None
        x.to(device=self.device)
        with set_grad(grad):
            y = self.model(**x.model_input)
        return _get_last_hidden_state(y)

    def set_annotations(
        self, docs: List[Doc], outputs: torch.Tensor, set_vector: bool = True
    ) -> None:
        """Assign the extracted features to the Doc.

        Args:
            docs: List of `spacy.Doc`.
            outputs: Output from `self.predict`.
            set_vector: If True, attach the vector to doc. This may harms the performance.
        """
        for i, doc in enumerate(docs):
            length = len(doc._.get(ATTRS.tokens))
            # Instead of assigning a tensor directly, assign `TensorWrapper`
            # so that trailing pipes can handle batch tensors efficiently.
            doc._.set(ATTRS.last_hidden_state, TensorWrapper(outputs, i, length))

            if set_vector:
                lh: torch.Tensor = doc._.get(ATTRS.last_hidden_state).get()
                doc_tensor = lh.new_zeros((len(doc), lh.shape[-1]))
                # TODO: Inefficient
                # TODO: Store the functionality into user_hooks after https://github.com/explosion/spaCy/issues/4439 has been released
                for i, a in enumerate(doc._.get(ATTRS.align)):
                    if self.max_length > 0:
                        a = [aa for aa in a if aa < len(lh)]
                    doc_tensor[i] += lh[a].sum(0)
                doc.tensor = doc_tensor
                doc.user_hooks["vector"] = get_doc_vector_via_tensor
                doc.user_span_hooks["vector"] = get_span_vector_via_tensor
                doc.user_token_hooks["vector"] = get_token_vector_via_tensor
                doc.user_hooks["similarity"] = get_similarity
                doc.user_span_hooks["similarity"] = get_similarity
                doc.user_token_hooks["similarity"] = get_similarity

    @property
    def freeze(self) -> bool:
        if self.cfg.get("freeze"):
            return True
        return False

    def update(self, docs: List[Doc], golds: List[GoldParse]):  # type: ignore
        """Simply forward `docs` in training mode."""
        if self.freeze:
            self.model.eval()
        else:
            self.model.train()
        y = self._apply_model(docs, not self.freeze)
        # `set_vector = False` because the tensor may not be necessary in updating.
        # The tensor is still available via doc._.transformers_last_hidden_state.
        self.set_annotations(docs, y, set_vector=False)


def _get_last_hidden_state(output: Tuple[Any]) -> torch.Tensor:
    # assumes output[0] is the last hidden state
    return output[0]


def get_doc_vector_via_tensor(doc) -> np.ndarray:
    return doc.tensor.sum(0).cpu().numpy()


def get_span_vector_via_tensor(span) -> np.ndarray:
    return span.doc.tensor[span.start : span.end].sum(0).cpu().numpy()


def get_token_vector_via_tensor(token) -> np.ndarray:
    return token.doc.tensor[token.i].cpu().numpy()



"""Defines KNP pipelines."""
import functools
import re
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Tuple

import spacy
from spacy.tokens import Doc, Span, Token
from spacy.util import filter_spans
from toolz import curry
from typing_extensions import Literal

from camphr.consts import JUMAN_LINES
from camphr.utils import get_juman_command

from .consts import KNP_USER_KEYS
from .noun_chunker import knp_noun_chunker

LOC2IOB = {"B": "B", "I": "I", "E": "I", "S": "B"}
Span.set_extension(JUMAN_LINES, default=None)
SKIP_TOKENS = {"@"}


TAG = "tag"
BUNSETSU = "bunsetsu"
MORPH = "morph"
L_KNP_OBJ = Literal["tag", "bunsetsu", "morph"]


def _take_juman_lines(n: int, juman_lines: List[str]) -> Tuple[List[str], List[str]]:
    lines = []
    count = 0
    for line in juman_lines:
        lines.append(line)
        head = line.split(" ")[0]
        if head not in SKIP_TOKENS:
            count += 1
        if count >= n:
            break
    return lines, juman_lines[len(lines) :]


@spacy.component("juman_sentencizer", requires=["doc.sents"])
def juman_sentencizer(doc: Doc) -> Doc:
    """Split juman_string for knp"""
    juman_lines = doc.user_data[JUMAN_LINES].strip("\n").split("\n")
    for sent in doc.sents:
        lines, juman_lines = _take_juman_lines(len(sent), juman_lines)
        sent._.set(JUMAN_LINES, "\n".join(lines) + "\n" + "EOS")
    return doc


# For spacy factory. see pyproject.toml
def juman_sentencizer_factory(*args, **kwargs):
    return juman_sentencizer


@spacy.component("knp", assigns=["doc.ents", "doc._.knp_tag_parent"])
class KNP:
    def __init__(
        self,
        knp_kwargs: Optional[Dict[str, str]] = None,
        preprocessor: Callable[[str], str] = None,
    ):
        import pyknp

        cmd = get_juman_command()
        assert cmd
        knp_kwargs = knp_kwargs or {}
        knp_kwargs.setdefault("jumancommand", cmd)

        self.knp = pyknp.KNP(**knp_kwargs)
        self.knp_kwargs = knp_kwargs

    @classmethod
    def from_nlp(cls, nlp, *args, **kwargs):
        return cls()

    def __call__(self, doc: Doc) -> Doc:
        for sent in doc.sents:
            blist = self.knp.parse_juman_result(sent._.get(JUMAN_LINES))
            mlist = blist.mrph_list()
            tlist = blist.tag_list()
            if len(mlist) != len(sent):
                mlist = _separate_mrph(mlist, sent)
            for label, comp in zip([blist, mlist, tlist], ["bunsetsu", "morph", "tag"]):
                sent._.set(getattr(KNP_USER_KEYS, comp).list_, label)
            for m, token in zip(mlist, sent):
                token._.set(KNP_USER_KEYS.morph.element, m)
        doc.ents = filter_spans(doc.ents + tuple(_extract_knp_ent(doc)))  # type: ignore
        doc.noun_chunks_iterator = knp_noun_chunker  # type: ignore
        # TODO: https://github.com/python/mypy/issues/3004
        return doc


def _separate_mrph(mlist: List, sent: Span) -> List:
    mm = []
    i = 0
    for m in mlist:
        if "形態素連結" in m.fstring:
            j = len(m.midasi)
            while j > 0:
                mm.append(m)
                j -= len(sent[i].text)
                i += 1
        elif sent[i].text == m.midasi:
            mm.append(m)
            i += 1
        else:
            raise ValueError(
                f"""Internal error occured
            Sentence: {sent.text}
            mlist : {[m.midasi for m in mlist]}
            tokens: {[t.text for t in sent]}
            diff  : {m.midasi}, {sent[i].text}
            """
            )
    return mm


@curry
@functools.lru_cache()
def token_to_knp_span(type_: str, token: Token) -> Span:
    """Returns the knp span containing the token."""
    assert type_ != MORPH
    for tag_or_bunsetsu in token.doc._.get(getattr(KNP_USER_KEYS, type_).spans):
        if token.i < tag_or_bunsetsu.end:
            return tag_or_bunsetsu
    raise ValueError("internal error")


@curry
@functools.lru_cache()
def get_knp_span(type_: str, span: Span) -> List[Span]:
    """Get knp tag or bunsetsu list"""
    assert type_ != MORPH

    # TODO: span._ => span.sent._
    knp_list = span._.get(getattr(KNP_USER_KEYS, type_).list_)
    if not knp_list:
        return []

    res = []
    i = span.start_char
    doc = span.doc
    for b in knp_list:
        j = i + len(b.midasi)
        bspan = doc.char_span(i, j)
        bspan._.set(getattr(KNP_USER_KEYS, type_).element, b)
        res.append(bspan)
        i = j
    return res


@functools.lru_cache()
def get_knp_element_id(elem) -> int:
    from pyknp import Morpheme, Bunsetsu, Tag

    for cls, attr in [(Morpheme, "mrph_id"), (Bunsetsu, "bnst_id"), (Tag, "tag_id")]:
        if isinstance(elem, cls):
            return getattr(elem, attr)
    raise ValueError(type(elem))


@curry
def get_all_knp_features_from_sents(
    type_: L_KNP_OBJ, feature: str, doc: Doc
) -> Iterator[Any]:
    """Helper for spacy.doc extension to get knp features from spans and concatenate them."""
    for sent in doc.sents:
        key = getattr(KNP_USER_KEYS, type_)
        yield from sent._.get(getattr(key, feature))


@curry
@functools.lru_cache()
def get_knp_parent(type_: L_KNP_OBJ, span: Span) -> Optional[Span]:
    tag_or_bunsetsu = span._.get(getattr(KNP_USER_KEYS, type_).element)
    if not tag_or_bunsetsu:
        return None
    p = tag_or_bunsetsu.parent
    if not p:
        return None
    spans = span.sent._.get(getattr(KNP_USER_KEYS, type_).spans)
    return spans[get_knp_element_id(p)]


@curry
@functools.lru_cache()
def get_knp_children(type_: L_KNP_OBJ, span: Span) -> List[Span]:
    tag_or_bunsetsu = span._.get(getattr(KNP_USER_KEYS, type_).element)
    if not tag_or_bunsetsu:
        return []
    children = tag_or_bunsetsu.children
    spans = span.sent._.get(getattr(KNP_USER_KEYS, type_).spans)
    return [spans[get_knp_element_id(child)] for child in children]


def _extract_knp_ent(doc: Doc) -> List[Span]:
    ents: List[Tuple[str, int, int]] = []
    for token in doc:
        ent_match = re.search(
            r"<NE:(\w+):(.*?)?>", token._.get(KNP_USER_KEYS.morph.element).fstring
        )
        if ent_match:
            ent, loc = ent_match.groups()
            iob = LOC2IOB[loc]
            if iob == "B":
                ents.append((ent, token.i, token.i + 1))
            else:
                last = ents[-1]
                ents[-1] = (last[0], last[1], token.i + 1)
    spacy_ents = _create_ents(doc, ents)
    return spacy_ents


def _create_ents(doc: Doc, ents: Iterable[Tuple[str, int, int]]) -> List[Span]:
    new_ents = []
    for text, start, end in ents:
        new_ents.append(Span(doc, start, end, label=text))
    return filter_spans(doc.ents + tuple(new_ents))


def _install_extensions():
    K = KNP_USER_KEYS
    Token.set_extension(K.morph.element, default=None, force=True)
    for k in ["bunsetsu", "tag"]:
        Token.set_extension(getattr(K.morph, k), getter=token_to_knp_span(k))
    for k in ["bunsetsu", "morph", "tag"]:
        for feature in ["element", "list_"]:
            key = getattr(getattr(K, k), feature)
            Span.set_extension(key, default=None, force=True)
    for k in ["bunsetsu", "morph", "tag"]:
        for feature in ["spans", "list_"]:
            key = getattr(getattr(K, k), feature)
            Doc.set_extension(key, getter=get_all_knp_features_from_sents(k, feature))
    for k in [BUNSETSU, TAG]:
        Span.set_extension(getattr(KNP_USER_KEYS, k).spans, getter=get_knp_span(k))
        Span.set_extension(getattr(KNP_USER_KEYS, k).parent, getter=get_knp_parent(k))
        Span.set_extension(
            getattr(KNP_USER_KEYS, k).children, getter=get_knp_children(k)
        )


_install_extensions()



"""The module torch_mixin defindes Language mixin for pytorch."""
import itertools
import logging
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union, cast

import spacy
import spacy.language
import srsly
import torch
from spacy.gold import GoldParse  # pylint: disable=no-name-in-module
from spacy.language import Language
from spacy.pipeline.pipes import Pipe
from spacy.scorer import Scorer
from spacy.tokens import Doc
from spacy.util import minibatch
from torch.optim.optimizer import Optimizer

from camphr.torch_utils import TorchPipe, get_loss_from_docs
from camphr.utils import get_defaults, get_requirements_line, import_attr

logger = logging.getLogger(__name__)


class TorchLanguage(Language):
    """spacy.Language for pytorch.

    This class manages all `TorchPipe` components for the sake of training.

    Examples:
        >>> nlp = TorchLanguage(Vocab(), meta={"lang": "en"})
        >>> nlp.add_pipe(foo_torch_pipe)
        >>> optim = nlp.resume_training()
        >>> nlp.update(docs, golds, optim)
    """

    LANG_FACTORY = "camphr_torch"

    def __init__(
        self,
        vocab=True,
        make_doc=True,
        max_length=10 ** 6,
        meta={},
        optimizer_config: Dict[str, Any] = {},
        **kwargs,
    ):
        meta = dict(meta)
        meta[
            "lang_factory"
        ] = self.LANG_FACTORY  # lang_factory is necessary when restoring.
        meta.setdefault("requirements", []).append(get_requirements_line())
        self.lang = meta.get("lang", "")
        self.Defaults = get_defaults(self.lang)
        self.optimizer_config = optimizer_config
        super().__init__(vocab, make_doc, max_length, meta=meta, **kwargs)

    def update(  # type: ignore
        self,
        docs: Sequence[Union[str, Doc]],
        golds: Sequence[Union[Dict[str, Any], GoldParse]],
        optimizer: Optimizer,
        verbose: bool = False,
    ):
        """Update `TorchPipe` models in pipeline."""
        _docs, _golds = self._format_docs_and_golds(docs, golds)
        self._update_pipes(_docs, _golds)
        self._update_params(_docs, optimizer, verbose)

    def _update_pipes(self, docs: Sequence[Doc], golds: Sequence[GoldParse]) -> None:
        for _, pipe in self.pipeline:
            pipe.update(docs, golds)

    def _update_params(
        self, docs: Sequence[Doc], optimizer: Optimizer, verbose: bool = False
    ):
        loss = get_loss_from_docs(docs)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        if verbose:
            logger.info(f"Loss: {loss.detach().item()}")

    def evaluate(  # type: ignore
        self,
        docs_golds: Sequence[Tuple[Union[str, Doc], Union[Dict[str, Any], GoldParse]]],
        batch_size: int = 16,
        scorer: Optional[Scorer] = None,
    ) -> Dict[str, Any]:
        """Overrided to compute loss"""
        if scorer is None:
            scorer = Scorer(pipeline=self.pipeline)
        loss = 0.0
        for batch in minibatch(docs_golds, size=batch_size):
            docs, golds = zip(*batch)
            docs, golds = self._format_docs_and_golds(docs, golds)  # type: ignore
            for _, pipe in self.pipeline:
                self._eval_pipe(pipe, docs, golds, batch_size=batch_size)
            loss += cast(float, get_loss_from_docs(docs).cpu().float().item())
            for doc, gold in zip(docs, golds):
                scorer.score(doc, gold)
        scores = scorer.scores
        scores["loss"] = loss
        return scores

    def _eval_pipe(
        self,
        pipe: Pipe,
        docs: Sequence[Doc],
        golds: Sequence[GoldParse],
        batch_size: int,
    ) -> Sequence[Doc]:
        if not hasattr(pipe, "pipe"):
            docs = spacy.language._pipe(docs, pipe, {})
        elif hasattr(pipe, "eval"):
            pipe.eval(docs, golds)  # type: ignore
        else:
            docs = list(pipe.pipe(docs, batch_size=batch_size))  # type: ignore
        return docs

    def resume_training(self, **kwargs) -> Optimizer:  # type: ignore
        """Gather all torch parameters in each `TorchPipe`s, and create an optimizer.

        Args:
            kwargs: passed to `self.make_optimizers`
        """
        assert hasattr(self, "pipeline")

        params = self.get_params()
        return self.create_optimizer(params, **kwargs)

    def get_params(self):
        return itertools.chain.from_iterable(
            pipe.optim_parameters()
            for _, pipe in self.pipeline
            if isinstance(pipe, TorchPipe)
        )

    def require_optimizer_config(self):
        assert isinstance(
            self.optimizer_config, dict
        ), "`self.optimizer_config` must be set."

    def create_optimizer(
        self, params: Iterable[Union[torch.Tensor, Dict[str, Any]]], **kwargs
    ) -> Optimizer:
        cls = import_attr(self.optimizer_config["class"])
        return cls(params, **self.optimizer_config.get("params", {}))

    @property
    def device(self) -> torch.device:
        if not hasattr(self, "_device"):
            self._device = torch.device("cpu")
            self.to(self._device)
        else:
            for pipe in self.get_torch_pipes():
                assert self._device.type == pipe.device.type
        return self._device

    def get_torch_pipes(self) -> List[TorchPipe]:
        return [pipe for _, pipe in self.pipeline if isinstance(pipe, TorchPipe)]

    def to(self, device: torch.device) -> bool:
        flag = False
        for pipe in self.get_torch_pipes():
            flag = True
            pipe.to(device)
        self._device = device
        return flag

    def to_disk(self, path: Union[str, Path], exclude=tuple(), disable=None):
        """Overrides Language.to_disk to save `nlp.lang` properly."""
        path = Path(path)
        super().to_disk(path, exclude, disable)
        meta = self.meta
        meta["lang"] = self.lang
        srsly.write_json(path / "meta.json", meta)


def get_torch_nlp(lang: str, **cfg):
    return TorchLanguage(meta={"lang": lang}, **cfg)



"""Convert KNP dependency parsing result to spacy format."""
from typing import Any, Dict, Iterable, Optional, List

import spacy
from spacy.symbols import (
    ADJ,
    ADP,
    ADV,
    AUX,
    CCONJ,
    DET,
    NOUN,
    NUM,
    PART,
    PRON,
    PUNCT,
    VERB,
)
from spacy.tokens import Doc, Span, Token

from camphr.pipelines.knp import KNP_USER_KEYS


@spacy.component("knp_dependency_parser", requires=("doc._.knp_tag_parent",))
def knp_dependency_parser(doc: Doc) -> Doc:
    tag_spans: Iterable[Span] = doc._.get(KNP_USER_KEYS.tag.spans)
    s = []
    for tag in tag_spans:
        for c in tag[1:]:
            c.head = tag[0]
            c.dep_ = _get_child_dep(c)
        parent: Optional[Span] = tag._.get(KNP_USER_KEYS.tag.parent)
        if parent is not None:
            tag[0].head = parent[0]
            tag[0].dep_ = _get_dep(tag[0])
        else:
            tag[0].head = tag[0]
            tag[0].dep_ = "ROOT"
        s.append(tag[0])
    s = _modify_head_punct(s)
    s = _modify_head_flat(s)
    s = _modify_head_conj(s)
    doc.is_parsed = True
    return doc


def knp_dependency_parser_factory(*args, **kwargs):
    return knp_dependency_parser


def _get_dep(tag: Token) -> str:
    ret = ({ADV: "advmod", CCONJ: "advmod", DET: "det"}).get(tag.pos)
    if ret:
        return ret
    elif tag.pos in {VERB, ADJ, PART}:
        if tag._.knp_morph_tag._.knp_tag_element.features.get("係", "") == "連格":
            return "acl"
        return "advcl"
    return _get_dep_noun(tag)


def _get_dep_noun(tag: Token) -> str:
    f: Dict[str, Any] = tag._.knp_morph_tag._.knp_tag_element.features
    if "係" not in f:
        return "dep"
    k = f["係"] if f["係"] != "未格" or "解析格" not in f else f["解析格"] + "格"
    x = {"隣": "nmod", "文節内": "compound", "ガ格": "nsubj", "ヲ格": "obj", "ガ２格": "dislocated"}
    if k in x:
        return x[k]
    elif k == "ノ格":
        if tag.head.pos in {VERB, ADJ}:
            return "nsubj"
        elif tag.pos in {DET, PRON}:
            tag.pos = DET
            return "det"
        else:
            return "nummod" if tag.pos == NUM else "nmod"
    elif "並列タイプ" in f:
        if tag.head.pos in {VERB, ADJ}:
            return "obl"
        else:
            return "conj"
    return "obl"


def _get_child_dep(tag: Token) -> str:
    p, pp = tag.pos, tag.head.pos
    if p == AUX:
        return "aux" if pp in {VERB, ADJ} else "cop"
    elif p == ADP:
        return "mark" if pp in {VERB, ADJ} else "case"
    elif p in {VERB, ADJ}:
        if pp == NOUN:
            tag.head.pos = VERB
        tag.pos = AUX
        return "aux"
    elif p == PART:
        return "mark"
    elif p == PUNCT:
        return "punct"
    else:
        return "clf" if pp == NUM else "flat"


def _modify_head_punct(heads: List[Token]) -> List[Token]:
    s = [t for t in heads]
    for i, t in enumerate(s):
        if t.pos != PUNCT:
            continue
        x = [u for u in t.rights]  # type: ignore
        if len(x) == 0:
            continue
        h = x[0]
        h.head = t.head
        h.dep_ = t.dep_
        x = x[1:] + [u for u in t.lefts]  # type: ignore
        x += [t, h] if h.dep_ == "ROOT" else [t]
        x += [u for u in s if u.head == t]
        for u in x:
            u.head = h
        t.dep_ = "punct"
        s[i] = h
    return s


def _modify_head_flat(heads: List[Token]) -> List[Token]:
    s = [t for t in heads]
    for i, t in enumerate(s):
        if not t.tag_.startswith("接頭辞"):
            continue
        x = [u for u in t.rights]  # type: ignore
        if len(x) == 0:
            continue
        h = x[0]
        if t.pos == NOUN and h.dep_ == "flat":
            d = "compound"
        elif t.pos == ADV and h.dep_ == "aux":
            d = "advmod"
            h.pos = VERB
        elif t.pos == PART and h.dep_ == "aux":
            d = "advmod"
            h.pos = ADJ
        else:
            continue
        h.head = t.head
        h.dep_ = t.dep_
        x = x[1:] + [u for u in t.lefts]  # type: ignore
        x += [t, h] if h.dep_ == "ROOT" else [t]
        x += [u for u in s if u.head == t]
        for u in x:
            u.head = h
        t.dep_ = d
        s[i] = h
    return s


def _modify_head_conj(heads: List[Token]) -> List[Token]:
    s = [t for t in heads]
    for t in s:
        while t.dep_ == "conj" and t.i < t.head.i:
            h = t.head
            t.head = h.head
            t.dep_ = h.dep_
            x = [h, t] if t.dep_ == "ROOT" else [h]
            x += [u for u in s if u.head == h and u.i < t.i]
            for u in x:
                u.head = t
            h.dep_ = "conj"
    return s



"""`models` defined `camphr.load`, camphr model loader."""
import functools
from collections import deque
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union, cast

import omegaconf
import spacy
import toolz
from omegaconf import OmegaConf
from spacy.language import Language
from spacy.pipeline import Pipe
from toolz import merge
from typing_extensions import Literal

from camphr.lang.torch import TorchLanguage
from camphr.ner_labels.utils import get_ner_labels
from camphr.pipelines.transformers.model import TRANSFORMERS_MODEL
from camphr.pipelines.transformers.ner import TRANSFORMERS_NER
from camphr.pipelines.transformers.seq_classification import (
    TRANSFORMERS_MULTILABEL_SEQ_CLASSIFIER,
    TRANSFORMERS_SEQ_CLASSIFIER,
)
from camphr.pipelines.transformers.tokenizer import TRANSFORMERS_TOKENIZER
from camphr.pipelines.transformers.utils import LABELS
from camphr.utils import get_labels, resolve_alias

__dir__ = Path(__file__).parent
_MODEL_CFG_DIR = __dir__ / "model_config"
_PREDEFINED_CONFS = {"knp": _MODEL_CFG_DIR / "knp.yml"}
CFG_SRC = Union[str, Dict[str, Any]]


def load(cfg: Union[str, CFG_SRC]) -> Language:
    if isinstance(cfg, str) and cfg in _PREDEFINED_CONFS:
        cfg = _PREDEFINED_CONFS[cfg].read_text()
    return create_model(cfg)


__all__ = ["load"]


@dataclass
class LangConfig(omegaconf.Config):
    name: str
    torch: bool
    optimizer: Dict[str, Any]
    kwargs: Optional[Dict[str, Any]]


@dataclass
class NLPConfig(omegaconf.Config):
    name: str
    lang: LangConfig
    pipeline: omegaconf.DictConfig
    task: Optional[Literal["ner", "textcat", "multilabel_textcat"]]
    labels: Optional[str]


def create_model(cfg: Union[NLPConfig, Any]) -> Language:
    if not isinstance(cfg, omegaconf.Config):
        cfg = OmegaConf.create(cfg)
    cfg = correct_model_config(cfg)
    nlp = create_lang(cfg.lang)
    for pipe in create_pipeline(nlp, cfg.pipeline):
        nlp.add_pipe(pipe)
    if cfg.name and isinstance(cfg.name, str):
        nlp._meta["name"] = cfg.name
    nlp._meta["config"] = OmegaConf.to_container(cfg.lang)
    return nlp


def create_lang(cfg: LangConfig) -> Language:
    kwargs = cfg.kwargs or {}
    kwargs = (
        OmegaConf.to_container(kwargs)
        if isinstance(kwargs, omegaconf.Config)
        else kwargs
    )
    if cfg.torch:
        kwargs["meta"] = merge(kwargs.get("meta", {}), {"lang": cfg.name})
        return TorchLanguage(True, optimizer_config=cfg.optimizer, **kwargs)
    return spacy.blank(cfg.name, **kwargs)


def create_pipeline(nlp: Language, cfg: omegaconf.DictConfig) -> List[Pipe]:
    if not isinstance(cfg, omegaconf.DictConfig):
        cfg = OmegaConf.create(cfg)

    pipes = []
    for name, pipe_config in cfg.items():
        pipe_config = OmegaConf.to_container(pipe_config or OmegaConf.create({}))
        pipes.append(nlp.create_pipe(name, config=pipe_config or dict()))
    return pipes


_ConfigParser = Callable[[NLPConfig], NLPConfig]


def correct_model_config(cfg: NLPConfig) -> NLPConfig:
    """Parse config. Complement missing informations, resolve aliases, etc."""
    PARSERS: List[_ConfigParser] = [
        resolve_alias(ALIASES),
        _add_pipes,
        _add_required_pipes,
        _align_pipeline,
        _correct_trf_pipeline,
        _resolve_label,
    ]
    return toolz.pipe(cfg, *PARSERS)


# Alias definition.
# For example, `pretrained` key is converted to `pipeline.transformers_model.trf_name_or_path`.
# The aliases in config is resolved by `resolved_alias`
ALIASES = {
    "pretrained": f"pipeline.{TRANSFORMERS_MODEL}.trf_name_or_path",
    "ner_label": f"pipeline.{TRANSFORMERS_NER}.labels",
    "textcat_label": f"pipeline.{TRANSFORMERS_SEQ_CLASSIFIER}.labels",
    "multitextcat_label": f"pipeline.{TRANSFORMERS_MULTILABEL_SEQ_CLASSIFIER}.labels",
    "optimizer": "lang.optimizer",
}


TRF_BASES = [TRANSFORMERS_TOKENIZER, TRANSFORMERS_MODEL]
TRF_TASKS = [
    TRANSFORMERS_SEQ_CLASSIFIER,
    TRANSFORMERS_NER,
    TRANSFORMERS_MULTILABEL_SEQ_CLASSIFIER,
]
TRF_PIPES = TRF_BASES + TRF_TASKS


# Mapping for pipeline requirements.
# {"pipename": [required pipe names in front]}
# e.g. `TRANSFORMERS_MODEL` requires `TRANSFORMERS_TOKENIZER` in front.
PIPELINE_ALIGNMENT = {
    TRANSFORMERS_MODEL: [TRANSFORMERS_TOKENIZER],
    TRANSFORMERS_NER: [TRANSFORMERS_MODEL],
    TRANSFORMERS_SEQ_CLASSIFIER: [TRANSFORMERS_MODEL],
    TRANSFORMERS_MULTILABEL_SEQ_CLASSIFIER: [TRANSFORMERS_MODEL],
}

TASK2PIPE = {
    "textcat": f"{TRANSFORMERS_SEQ_CLASSIFIER}",
    "ner": f"{TRANSFORMERS_NER}",
    "multilabel_textcat": f"{TRANSFORMERS_MULTILABEL_SEQ_CLASSIFIER}",
}


TASKS = {"textcat", "ner", "multilabel_textcat"}


def _add_pipes(cfg: NLPConfig) -> NLPConfig:
    if cfg.task in TASKS:
        assert cfg.labels, "`cfg.labels` required"
        cfg.pipeline = cfg.pipeline or OmegaConf.create({})
        # TODO https://github.com/microsoft/pyright/issues/671
        pipe = TASK2PIPE[cast(str, cfg.task)]
        prev = cfg.pipeline[pipe] or OmegaConf.create({})
        cfg.pipeline[pipe] = OmegaConf.merge(
            OmegaConf.create({"labels": cfg.labels}), prev
        )
    else:
        assert not cfg.labels, f"One of {TASKS} pipeline is required."
    return cfg


def _add_required_pipes(cfg: NLPConfig) -> NLPConfig:
    """assign required pipes defined in  `PIPELINE_ALIGNMENT`"""
    pipe_names = list(cfg.pipeline.keys())
    while True:
        prev_len = len(pipe_names)
        for k, v in PIPELINE_ALIGNMENT.items():
            for vv in v:
                if k in pipe_names and (vv not in pipe_names):
                    pipe_names.insert(pipe_names.index(k), vv)
        if prev_len == len(pipe_names):
            break
    cfg.pipeline = {k: cfg.pipeline.get(k, {}) for k in pipe_names}
    return cfg


def _align_pipeline(cfg: NLPConfig) -> NLPConfig:
    """align pipelines based on `PIPELINE_ALIGNMENT`"""
    pipe_names = sorted(cfg.pipeline.keys(), key=_pipe_cmp_key)
    cfg.pipeline = {k: cfg.pipeline.get(k, {}) for k in pipe_names}
    return cfg


@functools.cmp_to_key
def _pipe_cmp_key(x: str, y: str) -> int:
    """key function for `_align_pipeline`"""
    if _is_ancestor(x, y):
        return 1  # x comes after y
    if _is_ancestor(y, x):
        return -1  # y comes after x
    return 0  # no order relation to x and y


def _is_ancestor(child: str, ancestor: str) -> bool:
    if child not in PIPELINE_ALIGNMENT:
        return False
    parents = deque(PIPELINE_ALIGNMENT[child])
    while parents:
        p = parents.popleft()
        if p == ancestor:
            return True
        parents.extend(PIPELINE_ALIGNMENT.get(p, []))
    return False


def _correct_trf_pipeline(cfg: NLPConfig) -> NLPConfig:
    """Correct config of transformers pipeline

    Note:
        1. Complement `trf_name_or_path` for transformers pipelines.
        2. If there are trf pipe in pipeline, set lang.torch = true
    """
    cfg = _complement_trf_name(cfg)
    cfg = _correct_torch(cfg)
    return cfg


def _complement_trf_name(cfg: NLPConfig) -> NLPConfig:
    """For transformers pipeline.

    All transformers pipeline requires same `trf_name_or_path`, but it is a pain to write down.
    So if there is any transformers pipeline with `trf_name_or_path`, copy to other transformers pipes.
    """
    KEY = "trf_name_or_path"
    VAL = ""
    if not set(cfg.pipeline.keys()) & set(TRF_PIPES):
        return cfg
    for k, v in cfg.pipeline.items():
        VAL = v[KEY] or VAL
    if not VAL:
        raise ValueError(
            f"Invalid configuration. At least one of transformer's pipe needs `{KEY}`, but the configuration is:\n"
            + cfg.pipeline.pretty()
        )
    for k, v in cfg.pipeline.items():
        if k in TRF_PIPES and not v.get(KEY, None):
            v[KEY] = VAL
    return cfg


def _correct_torch(cfg: NLPConfig) -> NLPConfig:
    """If there is any transformers pipeline, set `torch` to True"""
    if set(cfg.pipeline) & set(TRF_PIPES):
        cfg.lang.torch = True
    return cfg


def _resolve_label(cfg: NLPConfig) -> NLPConfig:
    """Resolver for ner and sequence classification label config."""
    ner = cfg.pipeline[TRANSFORMERS_NER]
    if ner:
        ner[LABELS] = get_ner_labels(ner[LABELS])
    seq = cfg.pipeline[TRANSFORMERS_SEQ_CLASSIFIER]
    if seq:
        seq[LABELS] = get_labels(seq[LABELS])
    multiseq = cfg.pipeline[TRANSFORMERS_MULTILABEL_SEQ_CLASSIFIER]
    if multiseq:
        multiseq[LABELS] = get_labels(multiseq[LABELS])
    return cfg



from typing import Iterable, List, Optional, Tuple

import spacy
from spacy.tokens import Doc, Span, Token

from .consts import KNP_USER_KEYS

KNP_PARALLEL_NOUN_CHUNKS = "knp_parallel_noun_chunks"


def _install_extensions():
    Doc.set_extension(KNP_PARALLEL_NOUN_CHUNKS, default=None)


_install_extensions()


def knp_noun_chunker(doc: Doc) -> Iterable[Tuple[int, int, str]]:
    ret = []
    for taglist in _extract_noun_phrases(doc):
        last = _extract_content(taglist[-1])
        ret.append((taglist[0].start, last.end, "NP"))
    return ret


def get_parallel_noun_chunks(doc: Doc) -> List[List[Span]]:
    noun_phrases = list(_extract_noun_phrases(doc))
    last2idx = {taglist[-1]: i for i, taglist in enumerate(noun_phrases)}
    para_depends = {
        i: [_spans_to_span_without_last_aux(noun_phrases[i], "NP")]
        for i in range(len(noun_phrases))
    }
    for i, taglist in enumerate(noun_phrases):
        last = taglist[-1]
        if last._.get(KNP_USER_KEYS.tag.element).dpndtype == "P":
            parent = last2idx[last._.get(KNP_USER_KEYS.tag.parent)]
            para_depends[parent].extend(para_depends[i])
            del para_depends[i]
    return list([list(reversed(v)) for v in para_depends.values() if len(v) > 1])


@spacy.component(
    "knp_parallel_noun_chunker",
    requires=(f"span._.{KNP_USER_KEYS.tag.element}",),
    assigns=(f"doc._.{KNP_PARALLEL_NOUN_CHUNKS}",),
)
def knp_parallel_noun_chunker(doc: Doc) -> Doc:
    doc._.set(KNP_PARALLEL_NOUN_CHUNKS, get_parallel_noun_chunks(doc))
    return doc


def knp_parallel_noun_chunker_factory(*args, **kwargs):
    return knp_parallel_noun_chunker


def _spans_to_span_without_last_aux(spans: List[Span], label: str) -> Span:
    return _spans_to_span(spans[:-1] + [_extract_content(spans[-1])], label)


def _spans_to_span(spans: List[Span], label: str) -> Span:
    return Span(spans[0].doc, spans[0].start, spans[-1].end, label=label)


def _extract_noun_phrases(doc: Doc) -> Iterable[List[Span]]:
    seen = [False for _ in range(len(doc))]
    ret = []
    for tag in reversed(list(doc._.get(KNP_USER_KEYS.tag.spans))):
        if tag._.get(KNP_USER_KEYS.tag.element).features.get("体言"):
            taglist = _traverse_children(tag)
            i, j = _get_span(taglist)
            if seen[i]:
                # It is sufficient to check if the start has already been seen
                # because we are traversing the tags from the end of the doc.
                continue
            for k in range(i, j):
                seen[k] = True
            ret.append(taglist)
    return reversed(ret)


def _is_content(token: Token) -> bool:
    return "内容語>" in token._.get(KNP_USER_KEYS.morph.element).fstring


def _extract_content(tag: Span) -> Span:
    start = None
    end = None
    for token in tag:
        if _is_content(token):
            if start is None:
                start = token.i
            end = token.i + 1
        else:
            break
    assert start is not None and end is not None
    return Span(tag.doc, start, end)


def _traverse_children(
    tag: Span, _is_root: bool = True, _root: Optional[Span] = None
) -> List[Span]:
    """Traverse children except for `para` dependency.
    
    Args:
        tag: tag to be traversed
        _is_root: internally used parameter. Should not be changed outside this function.
        _root: internally used parameter. Should not be changed outside this function.
    """
    _root = _root or tag
    if (
        not _is_root
        and tag._.get(KNP_USER_KEYS.tag.element).dpndtype == "P"
        and tag._.get(KNP_USER_KEYS.tag.parent) == _root
    ):
        return []
    result = []
    for c in tag._.get(KNP_USER_KEYS.tag.children):
        result.extend(_traverse_children(c, False, _root))
    result.append(tag)
    return result


def _get_span(taglist: List[Span]) -> Tuple[int, int]:
    i = taglist[0].start
    j = taglist[-1].end
    return i, j



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['camphr',
 'camphr.cli',
 'camphr.lang',
 'camphr.lang.juman',
 'camphr.lang.mecab',
 'camphr.lang.sentencepiece',
 'camphr.ner_labels',
 'camphr.pipelines',
 'camphr.pipelines.knp',
 'camphr.pipelines.transformers']

package_data = \
{'': ['*'],
 'camphr': ['model_config/*'],
 'camphr.cli': ['conf/train/*', 'conf/train/example/*']}

install_requires = \
['catalogue>=1.0.0,<2.0.0',
 'dataclasses>=0.6,<0.7',
 'fire>=0.2.1,<1.0',
 'hydra-core>=0.11.3,<0.12.0',
 'hydra_colorlog>=0.1.4,<0.2.0',
 'more-itertools',
 'pyahocorasick>=1.4.0,<2.0.0',
 'pytokenizations>=0.4.8',
 'scikit-learn>=0.22,<0.24',
 'spacy>=2.2.3,<3.0.0',
 'toolz>=0.10.0,<0.11.0',
 'torch>=1.0,<2.0',
 'transformers>=2.10,<3.0',
 'typing-extensions>=3.7.4']

extras_require = \
{'all': ['allennlp>=1.0.0,<2.0.0',
         'unofficial-udify>=0.2.1,<0.3.0',
         'mojimoji>=0.0.11,<0.0.12',
         'unofficial-pyknp',
         'mecab-python3>=1.0,<1.1'],
 'allennlp': ['allennlp>=1.0.0,<2.0.0', 'boto3>=1.14,<2.0'],
 'juman': ['mojimoji>=0.0.11,<0.0.12', 'unofficial-pyknp'],
 'mecab': ['mecab-python3>=1.0,<1.1'],
 'udify': ['allennlp>=1.0.0,<2.0.0',
           'boto3>=1.14,<2.0',
           'unofficial-udify>=0.2.1,<0.3.0']}

entry_points = \
{'console_scripts': ['camphr = camphr.cli.__main__:main'],
 'spacy_factories': ['elmo = camphr.pipelines:Elmo.from_nlp',
                     'juman_sentencizer = '
                     'camphr.pipelines.knp:juman_sentencizer_factory',
                     'knp = camphr.pipelines.knp:KNP.from_nlp',
                     'knp_dependency_parser = '
                     'camphr.pipelines.knp.dependency_parser:knp_dependency_parser_factory',
                     'knp_parallel_noun_chunker = '
                     'camphr.pipelines.knp.noun_chunker:knp_parallel_noun_chunker_factory',
                     'multiple_regex_ruler = '
                     'camphr.pipelines:MultipleRegexRuler.from_nlp',
                     'pattern_searcher = camphr.pipelines:PatternSearcher',
                     'regex_ruler = camphr.pipelines:RegexRuler.from_nlp',
                     'transformers_model = '
                     'camphr.pipelines.transformers:TrfModel.from_nlp',
                     'transformers_ner = '
                     'camphr.pipelines.transformers:TrfForNamedEntityRecognition.from_nlp',
                     'transformers_sequece_classifier = '
                     'camphr.pipelines.transformers:TrfForSequenceClassification.from_nlp',
                     'transformers_sequence_classifier = '
                     'camphr.pipelines.transformers:TrfForSequenceClassification.from_nlp',
                     'transformers_tokenizer = '
                     'camphr.pipelines.transformers:TrfTokenizer.from_nlp',
                     'udify = camphr.pipelines:Udify.from_nlp'],
 'spacy_languages': ['camphr_torch = camphr.lang.torch:TorchLanguage',
                     'ja_juman = camphr.lang.juman:Japanese',
                     'ja_mecab = camphr.lang.mecab:Japanese',
                     'sentencepiece = '
                     'camphr.lang.sentencepiece:SentencePieceLang']}

setup_kwargs = {
    'name': 'camphr',
    'version': '0.6.2',
    'description': 'spaCy plugin for Transformers, Udify, Elmo, etc.',
    'long_description': '<p align="center"><img src="https://raw.githubusercontent.com/PKSHATechnology-Research/camphr/master/img/logoc.svg?sanitize=true" width="200" /></p>\n\n# Camphr - spaCy plugin for Transformers, Udify, Elmo, etc.\n\n[![Documentation Status](https://readthedocs.org/projects/camphr/badge/?version=latest)](https://camphr.readthedocs.io/en/latest/?badge=latest)\n[![Gitter](https://badges.gitter.im/camphr/community.svg)](https://gitter.im/camphr/community?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge)\n[![PyPI version](https://badge.fury.io/py/camphr.svg)](https://badge.fury.io/py/camphr)\n![test and publish](https://github.com/PKSHATechnology-Research/camphr/workflows/test%20and%20publish/badge.svg)\n![](https://github.com/PKSHATechnology-Research/camphr/workflows/test%20extras/badge.svg)\n![](https://github.com/PKSHATechnology-Research/camphr/workflows/test%20package/badge.svg)\n![daily-test](https://github.com/PKSHATechnology-Research/camphr/workflows/daily-test/badge.svg)\n\nCamphr is a *Natural Language Processing* library that helps in seamless integration for a wide variety of techniques from state-of-the-art to conventional ones.\nYou can use [Transformers](https://huggingface.co/transformers/) ,  [Udify](https://github.com/Hyperparticle/udify), [ELmo](https://allennlp.org/elmo), etc. on [spaCy](https://github.com/explosion/spaCy).\n\nCheck the [documentation](https://camphr.readthedocs.io/en/latest/) for more information.\n\n(For Japanese: https://qiita.com/tamurahey/items/53a1902625ccaac1bb2f)\n\n# Features\n\n- A [spaCy](https://github.com/explosion/spaCy) plugin - Easily integration for a wide variety of methods\n- [Transformers](https://huggingface.co/transformers/) with spaCy - Fine-tuning pretrained model with [Hydra](https://hydra.cc/). Embedding vector\n- [Udify](https://github.com/Hyperparticle/udify) - BERT based multitask model in 75 languages\n- [Elmo](https://allennlp.org/elmo) - Deep contextualized word representations\n- Rule base matching with Aho-Corasick, Regex\n- (for Japanese) KNP\n\n# License\n\nCamphr is licensed under [Apache 2.0](./LICENSE).\n\n',
    'author': 'tamuhey',
    'author_email': 'tamuhey@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/PKSHATechnology-Research/camphr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)
